﻿using System;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace TestProject
{
    /// <summary>
    /// 기정의 도형
    /// </summary>
    public static class PredefinedGeometry
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Constant
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 코사인 15도
        /// </summary>
        /// <remarks>PI / 12</remarks>
        private const float COS15 = 0.966f;

        /// <summary>
        /// 사인 15도
        /// </summary>
        private const float SIN15 = 0.259f;

        /// <summary>
        /// 코사인 45도
        /// </summary>
        /// <remarks>PI / 4</remarks>
        private const float COS45 = 0.707f;

        /// <summary>
        /// 사인 45도
        /// </summary>
        private const float SIN45 = 0.707f;

        /// <summary>
        /// 코사인 30도
        /// </summary>
        /// <remarks>PI / 6</remarks>
        private const float COS30 = 0.866f;

        /// <summary>
        /// 사인 30도
        /// </summary>
        private const float SIN30 = 0.500f;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 노드 패스 구하기 - GetNodePath(nodeShapeStyle, shapeOrientation)

        /// <summary>
        /// 노드 패스 구하기
        /// </summary>
        /// <param name="nodeShapeStyle">노드 도형 스타일</param>
        /// <param name="shapeOrientation">도형 방향</param>
        /// <returns>노드 패스</returns>
        public static GraphicsPath GetNodePath(NodeShapeStyle nodeShapeStyle, ShapeOrientation shapeOrientation)
        {
            GraphicsPath graphicsPath = new GraphicsPath();
            RectangleF   rectangle    = new RectangleF(0, 0, 1, 1);

            switch(nodeShapeStyle)
            {
                // 타원 도형들
                case NodeShapeStyle.Ellipse         :
                case NodeShapeStyle.Connector       :
                case NodeShapeStyle.Or              :
                case NodeShapeStyle.SummingJunction :

                    #region Ellipse/Connector/Or/SummingJunction

                    graphicsPath.AddEllipse(rectangle);

                    if(nodeShapeStyle == NodeShapeStyle.Or)
                    {
                        PointF centerPoint = PredefinedGeometry.GetCenterPoint(rectangle);

                        graphicsPath.AddLine(centerPoint.X, rectangle.Top, centerPoint.X, rectangle.Bottom);

                        graphicsPath.StartFigure();

                        graphicsPath.AddLine(rectangle.Left, centerPoint.Y, rectangle.Right, centerPoint.Y);
                    }
                    else if(nodeShapeStyle == NodeShapeStyle.SummingJunction)
                    {
                        PointF centerPointF = PredefinedGeometry.GetCenterPoint(rectangle);

                        float a = rectangle.Width  / 2 * COS45;
                        float b = rectangle.Height / 2 * SIN45;

                        graphicsPath.AddLine(centerPointF.X - a, centerPointF.Y - b, centerPointF.X + a, centerPointF.Y + b);

                        graphicsPath.StartFigure();

                        graphicsPath.AddLine(centerPointF.X + a, centerPointF.Y - b, centerPointF.X - a, centerPointF.Y + b);
                    }

                    break;

                    #endregion

                // 사각형 도형들
                case NodeShapeStyle.Rectangle         :
                case NodeShapeStyle.Process           :
                case NodeShapeStyle.PredefinedProcess :
                case NodeShapeStyle.InternalStorage   :

                    #region Rectangle/Process/PredefinedProcess/InternalStorage

                    graphicsPath.AddRectangle(rectangle);

                    if(nodeShapeStyle == NodeShapeStyle.PredefinedProcess)
                    {
                        graphicsPath.AddLine(rectangle.Left + rectangle.Width / 8, rectangle.Top, rectangle.Left + rectangle.Width / 8, rectangle.Bottom);

                        graphicsPath.StartFigure();

                        graphicsPath.AddLine(rectangle.Right - rectangle.Width / 8, rectangle.Top, rectangle.Right - rectangle.Width / 8, rectangle.Bottom);
                    }
                    else if(nodeShapeStyle == NodeShapeStyle.InternalStorage)
                    {
                        graphicsPath.AddLine(rectangle.Left + rectangle.Width / 8, rectangle.Top, rectangle.Left + rectangle.Width / 8, rectangle.Bottom);

                        graphicsPath.StartFigure();

                        graphicsPath.AddLine(rectangle.Left, rectangle.Top + rectangle.Height / 8, rectangle.Right, rectangle.Top + rectangle.Height / 8);
                    }

                    break;

                    #endregion

                case NodeShapeStyle.MultiDocument :

                    #region MultiDocument

                    MakeMultiDocumentPath(graphicsPath, rectangle);

                    break;

                    #endregion

                case NodeShapeStyle.Document :

                    #region Document

                    MakeDocumentPath(graphicsPath, rectangle);

                    break;

                    #endregion

                case NodeShapeStyle.PunchedTape :

                    #region PunchedTape

                    MakePunchedTapePath(graphicsPath, rectangle);

                    break;

                    #endregion

                case NodeShapeStyle.Delay :

                    #region Delay

                    if(rectangle.Width > 0 && rectangle.Height > 0)
                    {
                        graphicsPath.AddLine(rectangle.Left + rectangle.Width / 2, rectangle.Bottom, rectangle.Left                       , rectangle.Bottom);
                        graphicsPath.AddLine(rectangle.Left                       , rectangle.Bottom, rectangle.Left                       , rectangle.Top   );
                        graphicsPath.AddLine(rectangle.Left                       , rectangle.Top   , rectangle.Left + rectangle.Width / 2, rectangle.Top   );

                        graphicsPath.AddArc(rectangle.Left, rectangle.Top, rectangle.Width, rectangle.Height, 270, 180);
                    }

                    break;

                    #endregion

                case NodeShapeStyle.Termination :

                    #region Termination

                    if(rectangle.Width > 0 && rectangle.Height > 0)
                    {
                        graphicsPath.AddArc(rectangle.Left + rectangle.Width / 2, rectangle.Top, rectangle.Width / 2, rectangle.Height, -90, 180);
                        graphicsPath.AddArc(rectangle.Left                       , rectangle.Top, rectangle.Width / 2, rectangle.Height,  90, 180);

                        graphicsPath.CloseFigure();
                    }

                    break;

                    #endregion

                case NodeShapeStyle.MagneticDisk :

                    #region MagneticDisk

                    if(rectangle.Width > 0 && rectangle.Height > 0)
                    {
                        graphicsPath.FillMode = FillMode.Winding;

                        graphicsPath.AddLine(rectangle.Right, rectangle.Top + rectangle.Height / 8, rectangle.Right, rectangle.Bottom - rectangle.Height / 4);

                        graphicsPath.AddArc(rectangle.Left, rectangle.Bottom - rectangle.Height / 4, rectangle.Width, rectangle.Height / 4, 0, 180);

                        graphicsPath.AddLine(rectangle.Left, rectangle.Bottom - rectangle.Height / 4, rectangle.Left, rectangle.Top + rectangle.Height / 8);

                        graphicsPath.AddArc(rectangle.Left, rectangle.Top, rectangle.Width, rectangle.Height / 4, 180, 180);
                        graphicsPath.AddArc(rectangle.Left, rectangle.Top, rectangle.Width, rectangle.Height / 4, 0  , 180);
                    }

                    break;

                    #endregion

                case NodeShapeStyle.DirectAccessStorage :

                    #region DirectAccessStorage

                    if(rectangle.Width > 0 && rectangle.Height > 0)
                    {
                        graphicsPath.FillMode = FillMode.Winding;

                        graphicsPath.AddLine(rectangle.Left + rectangle.Width / 8, rectangle.Top, rectangle.Right - rectangle.Width / 4, rectangle.Top);

                        graphicsPath.AddArc(rectangle.Right - rectangle.Width / 4, rectangle.Top, rectangle.Width / 4, rectangle.Height, -90, 180);

                        graphicsPath.AddLine(rectangle.Right - rectangle.Width / 4, rectangle.Bottom, rectangle.Left + rectangle.Width / 8, rectangle.Bottom);

                        graphicsPath.AddArc(rectangle.Left, rectangle.Top, rectangle.Width / 4, rectangle.Height,  90, 180);
                        graphicsPath.AddArc(rectangle.Left, rectangle.Top, rectangle.Width / 4, rectangle.Height, -90, 180);
                    }

                    break;

                    #endregion

                case NodeShapeStyle.Display :

                    #region Display

                    if(rectangle.Width > 0 && rectangle.Height > 0)
                    {
                        graphicsPath.AddLine(rectangle.Right - rectangle.Width / 4, rectangle.Bottom                     , rectangle.Left  + rectangle.Width / 4, rectangle.Bottom                     );
                        graphicsPath.AddLine(rectangle.Left  + rectangle.Width / 4, rectangle.Bottom                     , rectangle.Left                        , rectangle.Top + rectangle.Height / 2);
                        graphicsPath.AddLine(rectangle.Left                        , rectangle.Top + rectangle.Height / 2, rectangle.Left  + rectangle.Width / 4, rectangle.Top                        );
                        graphicsPath.AddLine(rectangle.Left  + rectangle.Width / 4, rectangle.Top                        , rectangle.Right - rectangle.Width / 4, rectangle.Top                        );

                        graphicsPath.AddArc(rectangle.Right - rectangle.Width / 2, rectangle.Top, rectangle.Width / 2, rectangle.Height, 270, 180);
                    }

                    break;

                    #endregion

                case NodeShapeStyle.StoredData :

                    #region StoredData

                    if(rectangle.Width > 0 && rectangle.Height > 0)
                    {
                        graphicsPath.AddLine(rectangle.Left, rectangle.Top, rectangle.Right - rectangle.Width / 4, rectangle.Top);

                        graphicsPath.AddArc(rectangle.Right - rectangle.Width / 2, rectangle.Top, rectangle.Width / 2, rectangle.Height, 270, 180);

                        graphicsPath.AddLine(rectangle.Right - rectangle.Width / 4, rectangle.Bottom, rectangle.Left, rectangle.Bottom);

                        graphicsPath.AddArc(rectangle.Left - rectangle.Width / 4, rectangle.Top, rectangle.Width / 2, rectangle.Height, 90, -180);
                    }

                    break;

                    #endregion

                case NodeShapeStyle.OrGate :

                    #region OrGate

                    MakeOrGatePath(graphicsPath, rectangle);

                    break;

                    #endregion

                case NodeShapeStyle.Sort :

                    #region Sort

                    PointF[] polyPointArray = GetPolyPointArray(nodeShapeStyle, rectangle);

                    graphicsPath.AddLines(polyPointArray);

                    graphicsPath.CloseFigure();

                    graphicsPath.AddLine(polyPointArray[1], polyPointArray[3]);

                    break;

                    #endregion

                case NodeShapeStyle.RoundRect        :
                case NodeShapeStyle.AlternateProcess :

                    #region RoundRect/AlternateProcess

                    rectangle.Width  *= 16;
                    rectangle.Height *= 16;

                    MakeRoundedRectanglePath(graphicsPath, rectangle, new SizeF(2, 2));

                    Matrix matrix = new Matrix();

                    if(rectangle.Width != 0 && rectangle.Height != 0)
                    {
                        matrix.Scale(1 / rectangle.Width, 1 / rectangle.Height);
                    }

                    graphicsPath.Transform(matrix);

                    break;

                    #endregion

                case NodeShapeStyle.Triangle :
                case NodeShapeStyle.Extract  :

                    #region Triangle/Extract

                    graphicsPath.AddLine(rectangle.Left , rectangle.Bottom, rectangle.Right                      , rectangle.Bottom);
                    graphicsPath.AddLine(rectangle.Right, rectangle.Bottom, rectangle.Left + rectangle.Width / 2, rectangle.Top   );

                    graphicsPath.CloseFigure();

                    break;

                    #endregion

                case NodeShapeStyle.TriangleRectangle :

                    #region TriangleRectangle

                    graphicsPath.AddLine(rectangle.Left, rectangle.Top   , rectangle.Left , rectangle.Bottom);
                    graphicsPath.AddLine(rectangle.Left, rectangle.Bottom, rectangle.Right, rectangle.Bottom);

                    graphicsPath.CloseFigure();

                    break;

                    #endregion

                case NodeShapeStyle.Merge :

                    #region Merge

                    graphicsPath.AddLine(rectangle.Left , rectangle.Top, rectangle.Right                      , rectangle.Top   );
                    graphicsPath.AddLine(rectangle.Right, rectangle.Top, rectangle.Left + rectangle.Width / 2, rectangle.Bottom);

                    graphicsPath.CloseFigure();

                    break;

                    #endregion

                case NodeShapeStyle.Transport :

                    #region Transport

                    MakeTransportPath(graphicsPath, rectangle);

                    break;

                    #endregion

                default :

                    #region Default

                    graphicsPath.AddLines(GetPolyPointArray(nodeShapeStyle, rectangle));

                    graphicsPath.CloseFigure();

                    break;

                    #endregion
            }

            RectangleF pathRectangle;

            switch(nodeShapeStyle)
            {
                case NodeShapeStyle.Document      :
                case NodeShapeStyle.MultiDocument :
                case NodeShapeStyle.PunchedTape   :
                case NodeShapeStyle.Hexagon       :
                case NodeShapeStyle.Preparation   :
                case NodeShapeStyle.Pentagon      :
                case NodeShapeStyle.OrGate        :

                    GraphicsPath cloneGraphicsPath = (GraphicsPath)graphicsPath.Clone();

                    cloneGraphicsPath.Flatten(new Matrix(), 0.05F);

                    pathRectangle = cloneGraphicsPath.GetBounds();

                    break;

                default :

                    pathRectangle = graphicsPath.GetBounds();

                    break;
            }

            pathRectangle        = graphicsPath.GetBounds();
            pathRectangle.Width  = pathRectangle.Right;
            pathRectangle.Height = pathRectangle.Bottom;
            pathRectangle.X      = 0;
            pathRectangle.Y      = 0;

            Matrix targetMatrix = new Matrix();

            switch(shapeOrientation)
            {
                case ShapeOrientation.Degree0   : break;
                case ShapeOrientation.Degree90  : targetMatrix.RotateAt(-90 , PredefinedGeometry.GetCenterPoint(pathRectangle)); break;
                case ShapeOrientation.Degree180 : targetMatrix.RotateAt(-180, PredefinedGeometry.GetCenterPoint(pathRectangle)); break;
                case ShapeOrientation.Degree270 : targetMatrix.RotateAt(-270, PredefinedGeometry.GetCenterPoint(pathRectangle)); break;
            }

            graphicsPath.Transform(targetMatrix);

            return graphicsPath;
        }

        #endregion
        #region 노드 패스 구하기 - GetNodePath(nodeShapeStyle)

        /// <summary>
        /// 노드 패스 구하기
        /// </summary>
        /// <param name="nodeShapeStyle">노드 도형 스타일</param>
        /// <returns>노드 패스</returns>
        public static GraphicsPath GetNodePath(NodeShapeStyle nodeShapeStyle)
        {
            return GetNodePath(nodeShapeStyle, ShapeOrientation.Degree0);
        }

        #endregion

        #region 링크 화살표 스타일 패스 구하기 - GetLinkArrowStyle(linkArrowStyle, width, height)

        /// <summary>
        /// 링크 화살표 스타일 패스 구하기
        /// </summary>
        /// <param name="linkArrowStyle">링크 화살표 스타일</param>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>링크 화살표 스타일 패스</returns>
        public static GraphicsPath GetLinkArrowStylePath(LinkArrowStyle linkArrowStyle, float width, float height)
        {
            switch(linkArrowStyle)
            {
                case LinkArrowStyle.Circle          : return GetCirclePath(width, height); 
                case LinkArrowStyle.Arrow           : return GetArrowPath(width, height);
                case LinkArrowStyle.Arrow2          : return GetArrow2Path(width, height);
                case LinkArrowStyle.HalfArrow       : return GetHalfArrowPath(width, height);
                case LinkArrowStyle.OpenedArrow     : return GetOpenedArrowPath(width, height);
                case LinkArrowStyle.Losange         : return GetLosangePath(width, height);
                case LinkArrowStyle.Fork            : return GetForkPath(width, height);
                case LinkArrowStyle.ClosedFork      : return GetClosedForkPath(width, height);
                case LinkArrowStyle.Many            : return GetManyPath(width, height);
                case LinkArrowStyle.ManyOptional    : return GetManyOptionalPath(width, height);
                case LinkArrowStyle.OneOrMany       : return GetOneOrManyPath(width, height);
                case LinkArrowStyle.One             : return GetOnePath(width, height);
                case LinkArrowStyle.OneOptional     : return GetOneOptionalPath(width, height);
                case LinkArrowStyle.OpenedHalfArrow : return GetOpenedHalfArrowPath(width, height);
                default                             : return null;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region Or Gate 패스 만들기 - MakeOrGatePath(graphicsPath, rectangle)

        /// <summary>
        /// Or Gate 패스 만들기
        /// </summary>
        /// <param name="graphicsPath">그래픽스 객체</param>
        /// <param name="rectangle">사각형</param>
        private static void MakeOrGatePath(GraphicsPath graphicsPath, RectangleF rectangle)
        {
            PointF point0 = new PointF(rectangle.Left                                        , rectangle.Top                                          );
            PointF point1 = new PointF(rectangle.Left + rectangle.Width + rectangle.Width / 3, rectangle.Top                                          );
            PointF point2 = new PointF(rectangle.Left + rectangle.Width + rectangle.Width / 3, rectangle.Top + rectangle.Height                       );
            PointF point3 = new PointF(rectangle.Left                                        , rectangle.Top + rectangle.Height                       );
            PointF point4 = new PointF(rectangle.Left + 8 * rectangle.Width / 15             , rectangle.Top + rectangle.Height - rectangle.Height / 6);
            PointF point5 = new PointF(rectangle.Left + 8 * rectangle.Width / 15             , rectangle.Top + rectangle.Height / 6                   );

            graphicsPath.AddBezier(point0, point1, point2, point3);
            graphicsPath.AddBezier(point3, point4, point5, point0);

            graphicsPath.CloseFigure();
        }

        #endregion
        #region Document 패스 만들기 - MakeDocumentPath(graphicsPath, rectangle)

        /// <summary>
        /// Document 패스 만들기
        /// </summary>
        /// <param name="graphicsPath">그래픽스 패스</param>
        /// <param name="rectangle">사각형</param>
        private static void MakeDocumentPath(GraphicsPath graphicsPath, RectangleF rectangle)
        {
            graphicsPath.AddLine(rectangle.Left , rectangle.Bottom - rectangle.Height / 16, rectangle.Left , rectangle.Top                           );
            graphicsPath.AddLine(rectangle.Left , rectangle.Top                           , rectangle.Right, rectangle.Top                           );
            graphicsPath.AddLine(rectangle.Right, rectangle.Top                           , rectangle.Right, rectangle.Bottom - rectangle.Height / 16);

            PointF point1 = new PointF(rectangle.Right                      , rectangle.Bottom - rectangle.Height / 16);
            PointF point2 = new PointF(rectangle.Right - rectangle.Width / 4, rectangle.Bottom - rectangle.Height / 4 );
            PointF point3 = new PointF(rectangle.Left  + rectangle.Width / 4, rectangle.Bottom + rectangle.Height / 8 );
            PointF point4 = new PointF(rectangle.Left                       , rectangle.Bottom - rectangle.Height / 16);

            graphicsPath.AddBezier(point1, point2, point3, point4);

            graphicsPath.CloseFigure();
        }

        #endregion
        #region Multi Document 패스 만들기 - MakeMultiDocumentPath(graphicsPath, rectangle)

        /// <summary>
        /// Multi Document 패스 만들기
        /// </summary>
        /// <param name="graphicsPath">그래픽스 패스</param>
        /// <param name="rectangle">사각형</param>
        static void MakeMultiDocumentPath(GraphicsPath graphicsPath, RectangleF rectangle)
        {
            PointF[] pointArray = new PointF[15];

            pointArray[0 ].X = rectangle.Left;
            pointArray[0 ].Y = rectangle.Bottom - rectangle.Height / 16;
            pointArray[1 ].X = rectangle.Left;
            pointArray[1 ].Y = rectangle.Top  + rectangle.Height / 4;
            pointArray[2 ].X = rectangle.Left + rectangle.Width  / 8;
            pointArray[2 ].Y = rectangle.Top  + rectangle.Height / 4;
            pointArray[3 ].X = rectangle.Left + rectangle.Width  / 8;
            pointArray[3 ].Y = rectangle.Top  + rectangle.Height / 4 - rectangle.Height / 8;
            pointArray[4 ].X = rectangle.Left + rectangle.Width  / 4;
            pointArray[4 ].Y = rectangle.Top  + rectangle.Height / 4 - rectangle.Height / 8;
            pointArray[5 ].X = rectangle.Left + rectangle.Width  / 4;
            pointArray[5 ].Y = rectangle.Top;
            pointArray[6 ].X = rectangle.Right;
            pointArray[6 ].Y = rectangle.Top;
            pointArray[7 ].X = rectangle.Right;
            pointArray[7 ].Y = rectangle.Bottom - rectangle.Height / 4;
            pointArray[8 ].X = rectangle.Right  - rectangle.Width  / 8;
            pointArray[8 ].Y = rectangle.Bottom - rectangle.Height / 4;
            pointArray[9 ].X = rectangle.Right  - rectangle.Width  / 8;
            pointArray[9 ].Y = rectangle.Bottom - rectangle.Height / 8;
            pointArray[10].X = rectangle.Right  - rectangle.Width  / 4;
            pointArray[10].Y = rectangle.Bottom - rectangle.Height / 8;
            pointArray[11].X = rectangle.Right  - rectangle.Width  / 4;
            pointArray[11].Y = rectangle.Bottom - 3 * rectangle.Height / 64;
            pointArray[12].X = rectangle.Right  - rectangle.Width / 4 - 3 * rectangle.Width / 16;
            pointArray[12].Y = rectangle.Bottom - 3 * rectangle.Height / 32;
            pointArray[13].X = rectangle.Left   + 3 * rectangle.Width  / 16;
            pointArray[13].Y = rectangle.Bottom;
            pointArray[14].X = rectangle.Left;
            pointArray[14].Y = rectangle.Bottom - 3 * rectangle.Height / 64;

            pointArray[12].Y -= 3 * rectangle.Height / 32;
            pointArray[13].Y += 3 * rectangle.Height / 32;

            PointF[] pointArray2 = new PointF[12];

            Array.Copy(pointArray, pointArray2, 12);

            graphicsPath.AddLines(pointArray2);

            graphicsPath.AddBezier(pointArray[11], pointArray[12], pointArray[13], pointArray[14]);

            graphicsPath.StartFigure();

            graphicsPath.AddLine(pointArray[2].X, pointArray[2].Y, pointArray[10].X, pointArray[2].Y);

            graphicsPath.StartFigure();

            graphicsPath.AddLine(pointArray[10].X, pointArray[2].Y, pointArray[10].X, pointArray[10].Y);

            graphicsPath.StartFigure();

            graphicsPath.AddLine(pointArray[4].X, pointArray[4].Y, pointArray[8].X, pointArray[4].Y);

            graphicsPath.StartFigure();

            graphicsPath.AddLine(pointArray[8].X, pointArray[4].Y, pointArray[8].X, pointArray[8].Y);
        }

        #endregion
        #region Puched Tape 패스 만들기 - MakePunchedTapePath(graphicsPath, rectangle)

        /// <summary>
        /// Puched Tape 패스 만들기
        /// </summary>
        /// <param name="graphicsPath">그래픽스 패스</param>
        /// <param name="rectangle">사각형</param>
        private static void MakePunchedTapePath(GraphicsPath graphicsPath, RectangleF rectangle)
        {
            PointF point1 = new PointF(rectangle.Left                       , rectangle.Top    + rectangle.Height / 16);
            PointF point2 = new PointF(rectangle.Left  + rectangle.Width / 4, rectangle.Top    + rectangle.Height / 4 );
            PointF point3 = new PointF(rectangle.Right - rectangle.Width / 4, rectangle.Top    - rectangle.Height / 8 );
            PointF point4 = new PointF(rectangle.Right                      , rectangle.Top    + rectangle.Height / 16);
            PointF point5 = new PointF(rectangle.Right                      , rectangle.Bottom - rectangle.Height / 16);
            PointF point6 = new PointF(rectangle.Right - rectangle.Width / 4, rectangle.Bottom - rectangle.Height / 4 );
            PointF point7 = new PointF(rectangle.Left  + rectangle.Width / 4, rectangle.Bottom + rectangle.Height / 8 );
            PointF point8 = new PointF(rectangle.Left                       , rectangle.Bottom - rectangle.Height / 16);

            graphicsPath.AddLine(rectangle.Left, rectangle.Bottom - rectangle.Height / 16, rectangle.Left, rectangle.Top + rectangle.Height / 16);

            graphicsPath.AddBezier(point1, point2, point3, point4);

            graphicsPath.AddLine(rectangle.Right, rectangle.Top + rectangle.Height / 16, rectangle.Right, rectangle.Bottom - rectangle.Height / 16);

            graphicsPath.AddBezier(point5, point6, point7, point8);
        }

        #endregion
        #region Transport 패스 만들기 - MakeTransportPath(graphicsPath, rectangle)

        /// <summary>
        /// Transport 패스 만들기
        /// </summary>
        /// <param name="graphicsPath">그래픽스 패스</param>
        /// <param name="rectangle">사각형</param>
        static void MakeTransportPath(GraphicsPath graphicsPath, RectangleF rectangle)
        {
            PointF[] pointArray = new PointF[10];

            pointArray[0].X = rectangle.Left;
            pointArray[0].Y = rectangle.Top    + rectangle.Height / 2;
            pointArray[1].X = rectangle.Left   + rectangle.Width  / 4;
            pointArray[1].Y = rectangle.Top;
            pointArray[2].X = rectangle.Left   + rectangle.Width  / 4;
            pointArray[2].Y = rectangle.Top    + rectangle.Height / 4;
            pointArray[3].X = rectangle.Right  - rectangle.Width  / 4;
            pointArray[3].Y = rectangle.Top    + rectangle.Height / 4;
            pointArray[4].X = rectangle.Right  - rectangle.Width  / 4;
            pointArray[4].Y = rectangle.Top;
            pointArray[5].X = rectangle.Right;
            pointArray[5].Y = rectangle.Top    + rectangle.Height / 2;
            pointArray[6].X = rectangle.Right  - rectangle.Width  / 4;
            pointArray[6].Y = rectangle.Bottom;
            pointArray[7].X = rectangle.Right  - rectangle.Width  / 4;
            pointArray[7].Y = rectangle.Bottom - rectangle.Height / 4;
            pointArray[8].X = rectangle.Left   + rectangle.Width  / 4;
            pointArray[8].Y = rectangle.Bottom - rectangle.Height / 4;
            pointArray[9].X = rectangle.Left   + rectangle.Width  / 4;
            pointArray[9].Y = rectangle.Bottom;

            graphicsPath.AddLines(pointArray);

            graphicsPath.CloseFigure();
        }

        #endregion
        #region Rounded Rectangle 패스 만들기 - MakeRoundedRectanglePath(graphicsPath, rectangle, size)

        /// <summary>
        /// Rounded Rectangle 패스 만들기
        /// </summary>
        /// <param name="graphicsPath">그래픽스 패스</param>
        /// <param name="rectangle">사각형 크기</param>
        /// <param name="size">크기</param>
        static void MakeRoundedRectanglePath(GraphicsPath graphicsPath, RectangleF rectangle, SizeF size)
        {
            if(rectangle.Width > 0 && rectangle.Height > 0)
            {
                graphicsPath.AddArc(rectangle.Right - size.Width, rectangle.Top                 , size.Width, size.Height, 270, 90);
                graphicsPath.AddArc(rectangle.Right - size.Width, rectangle.Bottom - size.Height, size.Width, size.Height, 0  , 90);
                graphicsPath.AddArc(rectangle.Left              , rectangle.Bottom - size.Height, size.Width, size.Height, 90 , 90);
                graphicsPath.AddArc(rectangle.Left              , rectangle.Top                 , size.Width, size.Height, 180, 90);

                graphicsPath.CloseFigure();
            }
        }

        #endregion

        #region 다수 포인트 배열 구하기 - GetPolyPointArray(nodeShapeStyle, rectangle)

        /// <summary>
        /// 다수 포인트 배열 구하기
        /// </summary>
        /// <param name="nodeShapeStyle">노드 도형 스타일</param>
        /// <param name="rectangle">사각형</param>
        /// <returns>다수 포인트 배열</returns>
        private static PointF[] GetPolyPointArray(NodeShapeStyle nodeShapeStyle, RectangleF rectangle)
        {
            float    angle;
            float    theta;
            float    slice;
            float    degreeToRadian;
            PointF[] pointArray;

            switch(nodeShapeStyle)
            {
                case NodeShapeStyle.Sort        :
                case NodeShapeStyle.Losange     :
                case NodeShapeStyle.Decision    :
                case NodeShapeStyle.Hexagon     :
                case NodeShapeStyle.Preparation :
                case NodeShapeStyle.Pentagon    :
                case NodeShapeStyle.Octogon     :

                    angle = 0;

                    int pointCount = 0;

                    switch(nodeShapeStyle)
                    {
                        case NodeShapeStyle.Sort     :
                        case NodeShapeStyle.Losange  :
                        case NodeShapeStyle.Decision :

                            pointCount = 4;
                            pointArray = new PointF[pointCount];

                            break;

                        case NodeShapeStyle.Pentagon :

                            pointCount = 5;
                            pointArray = new PointF[pointCount];
                            angle      = 90;

                            break;

                        case NodeShapeStyle.Hexagon     :
                        case NodeShapeStyle.Preparation :

                            pointCount = 6;
                            pointArray = new PointF[pointCount];
                            angle      = 90;

                            break;

                        case NodeShapeStyle.Octogon :

                            pointCount = 8;
                            pointArray = new PointF[pointCount];

                            break;

                        default :

                            pointCount = 0;
                            pointArray = null;

                            break;
                    }

                    degreeToRadian = (float)(3.1416 / 180);

                    slice = 360 / pointCount;

                    for(int i = 0; i < pointCount; i++)
                    {
                        theta = angle * degreeToRadian;

                        pointArray[i].X = rectangle.Left + rectangle.Width  / 2 + rectangle.Width  / 2 * (float)Math.Sin(theta);
                        pointArray[i].Y = rectangle.Top  + rectangle.Height / 2 + rectangle.Height / 2 * (float)Math.Cos(theta);

                        angle += slice;
                    }

                    break;

                case NodeShapeStyle.ManualOperation :

                    pointArray = new PointF[4];

                    pointArray[0].X = rectangle.Left;
                    pointArray[0].Y = rectangle.Top;
                    pointArray[1].X = rectangle.Left + rectangle.Width;
                    pointArray[1].Y = rectangle.Top  + rectangle.Height / 5;
                    pointArray[2].X = rectangle.Left + rectangle.Width;
                    pointArray[2].Y = rectangle.Top  + rectangle.Height - rectangle.Height / 5;
                    pointArray[3].X = rectangle.Left;
                    pointArray[3].Y = rectangle.Top  + rectangle.Height;

                    break;

                case NodeShapeStyle.ManualInput :

                    pointArray = new PointF[4];

                    pointArray[0].X = rectangle.Left;
                    pointArray[0].Y = rectangle.Top  + rectangle.Height / 5;
                    pointArray[1].X = rectangle.Left + rectangle.Width;
                    pointArray[1].Y = rectangle.Top;
                    pointArray[2].X = rectangle.Left + rectangle.Width;
                    pointArray[2].Y = rectangle.Top  + rectangle.Height;
                    pointArray[3].X = rectangle.Left;
                    pointArray[3].Y = rectangle.Top  + rectangle.Height;

                    break;

                case NodeShapeStyle.Collate :

                    pointArray = new PointF[5];

                    pointArray[0].X = rectangle.Left;
                    pointArray[0].Y = rectangle.Top;
                    pointArray[1].X = rectangle.Left + rectangle.Width;
                    pointArray[1].Y = rectangle.Top;
                    pointArray[2].X = rectangle.Left;
                    pointArray[2].Y = rectangle.Top  + rectangle.Height;
                    pointArray[3].X = rectangle.Left + rectangle.Width;
                    pointArray[3].Y = rectangle.Top  + rectangle.Height;
                    pointArray[4].X = rectangle.Left;
                    pointArray[4].Y = rectangle.Top;

                    break;

                case NodeShapeStyle.Data :

                    pointArray = new PointF[4];

                    pointArray[0].X = rectangle.Left + rectangle.Width / 4;
                    pointArray[0].Y = rectangle.Top;
                    pointArray[1].X = rectangle.Left + rectangle.Width;
                    pointArray[1].Y = rectangle.Top;
                    pointArray[2].X = rectangle.Left + rectangle.Width - rectangle.Width / 4;
                    pointArray[2].Y = rectangle.Top  + rectangle.Height;
                    pointArray[3].X = rectangle.Left;
                    pointArray[3].Y = rectangle.Top  + rectangle.Height;

                    break;

                case NodeShapeStyle.SequentialAccessStorage :

                    pointArray     = new PointF[24];
                    degreeToRadian = (float)(3.1416 / 180);
                    angle          = -45;

                    for(int i = 0; i < 22; i++)
                    {
                        theta = angle * degreeToRadian;

                        pointArray[i].X = rectangle.Left + rectangle.Width  / 2 + rectangle.Width  / 2 * (float)Math.Cos(theta);
                        pointArray[i].Y = rectangle.Top  + rectangle.Height / 2 - rectangle.Height / 2 * (float)Math.Sin(theta);

                        angle += 15;
                    }

                    pointArray[22].X = rectangle.Left + rectangle.Width;
                    pointArray[22].Y = pointArray[21].Y;
                    pointArray[23].X = rectangle.Left + rectangle.Width;
                    pointArray[23].Y = pointArray[0].Y;

                    break;

                case NodeShapeStyle.OffPageConnection :

                    pointArray = new PointF[5];

                    pointArray[0].X = rectangle.Left + rectangle.Width;
                    pointArray[0].Y = rectangle.Top  + rectangle.Height / 2;
                    pointArray[1].X = rectangle.Left + rectangle.Width  / 2;
                    pointArray[1].Y = rectangle.Top  + rectangle.Height;
                    pointArray[2].X = rectangle.Left;
                    pointArray[2].Y = rectangle.Top  + rectangle.Height;
                    pointArray[3].X = rectangle.Left;
                    pointArray[3].Y = rectangle.Top;
                    pointArray[4].X = rectangle.Left + rectangle.Width / 2;
                    pointArray[4].Y = rectangle.Top;

                    break;

                case NodeShapeStyle.ProcessISO9000 :

                    pointArray = new PointF[6];

                    pointArray[0].X = rectangle.Left + rectangle.Width;
                    pointArray[0].Y = rectangle.Top  + rectangle.Height / 2;
                    pointArray[1].X = rectangle.Left + rectangle.Width - rectangle.Width / 4;
                    pointArray[1].Y = rectangle.Top  + rectangle.Height;
                    pointArray[2].X = rectangle.Left;
                    pointArray[2].Y = rectangle.Top  + rectangle.Height;
                    pointArray[3].X = rectangle.Left + rectangle.Width  / 4;
                    pointArray[3].Y = rectangle.Top  + rectangle.Height / 2;
                    pointArray[4].X = rectangle.Left;
                    pointArray[4].Y = rectangle.Top;
                    pointArray[5].X = rectangle.Left + rectangle.Width - rectangle.Width / 4;
                    pointArray[5].Y = rectangle.Top;

                    break;

                case NodeShapeStyle.Card :

                    pointArray = new PointF[5];

                    pointArray[0].X = rectangle.Left + rectangle.Width / 6;
                    pointArray[0].Y = rectangle.Top;
                    pointArray[1].X = rectangle.Left + rectangle.Width;
                    pointArray[1].Y = rectangle.Top;
                    pointArray[2].X = rectangle.Left + rectangle.Width;
                    pointArray[2].Y = rectangle.Top  + rectangle.Height;
                    pointArray[3].X = rectangle.Left;
                    pointArray[3].Y = rectangle.Top  + rectangle.Height;
                    pointArray[4].X = rectangle.Left;
                    pointArray[4].Y = rectangle.Top  + rectangle.Height / 6;

                    break;

                default :

                    pointArray = null;

                    break;
            }

            return pointArray;
        }

        #endregion


        #region 원 패스 구하기 - GetCirclePath(width, height)

        /// <summary>
        /// 원 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>원 패스</returns>
        private static GraphicsPath GetCirclePath(float width, float height)
        {
            GraphicsPath graphicsPath = new GraphicsPath();
            RectangleF   rectangle    = new RectangleF(0, 0, width, height);

            graphicsPath.AddEllipse(rectangle);

            return graphicsPath;
        }

        #endregion
        #region 화살표 패스 구하기 - GetArrowPath(width, height)

        /// <summary>
        /// 화살표 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>화살표 패스</returns>
        private static GraphicsPath GetArrowPath(float width, float height)
        {
            GraphicsPath graphicsPath = new GraphicsPath();
            PointF[]     pointArray   = { new PointF(0, 0), new PointF(width, height/2), new PointF(0, height) };

            graphicsPath.AddPolygon(pointArray);

            return graphicsPath;
        }

        #endregion
        #region 화살표 2 패스 구하기 - GetArrow2Path(width, height)

        /// <summary>
        /// 화살표 2 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>화살표 2 패스</returns>
        private static GraphicsPath GetArrow2Path(float width, float height)
        {
            GraphicsPath graphicsPath = new GraphicsPath();
            PointF[]     pointArray   = { new PointF(width / 4, height / 2), new PointF(0, 0), new PointF(width, height / 2), new PointF(0, height) };

            graphicsPath.AddPolygon(pointArray);

            return graphicsPath;
        }   		

        #endregion
        #region 개방형 화살표 패스 구하기 - GetOpenedArrowPath(width, height)

        /// <summary>
        /// 개방형 화살표 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>개방형 화살표 패스</returns>
        private static GraphicsPath GetOpenedArrowPath(float width, float height)
        {
            GraphicsPath graphicsPath = new GraphicsPath();

            graphicsPath.AddLine(new PointF(width, height / 2), new PointF(0, 0     ));
            graphicsPath.AddLine(new PointF(width, height / 2), new PointF(0, height));

            return graphicsPath;
        }

        #endregion
        #region 마름모 패스 구하기 - GetLosangePath(width, height)

        /// <summary>
        /// 마름모 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>마름모 패스</returns>
        private static GraphicsPath GetLosangePath(float width, float height)
        {
            GraphicsPath graphicsPath = new GraphicsPath();
            PointF[]     pointArray   = { new PointF(0, height / 2), new PointF(width / 2, height), new PointF(width, height / 2), new PointF(width / 2, 0) };

            graphicsPath.AddPolygon(pointArray);

            return graphicsPath;
        }   		

        #endregion
        #region 포크 패스 구하기 - GetForkPath(width, height)

        /// <summary>
        /// 포크 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>포크 패스</returns>
        private static GraphicsPath GetForkPath(float width, float height)
        {
            GraphicsPath graphicsPath = new GraphicsPath();
            PointF[]     pointArray   = { new PointF(0, height / 2), new PointF(width, 0), new PointF(width, height) };

            graphicsPath.AddPolygon(pointArray);

            return graphicsPath;
        }   		

        #endregion
        #region 폐쇄형 포크 패스 구하기 - GetClosedForkPath(width, height)

        /// <summary>
        /// 폐쇄형 포크 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>폐쇄형 포크 패스</returns>
        private static GraphicsPath GetClosedForkPath(float width, float height)
        {
            GraphicsPath graphicsPath = new GraphicsPath();
            PointF[]     pointArray   = { new PointF(0, height / 2), new PointF(width, 0), new PointF(width, height) };

            graphicsPath.AddPolygon(pointArray);

            graphicsPath.AddLine(new PointF(0, height), new PointF(0, 0));

            return graphicsPath;
        }   		

        #endregion
        #region Many 패스 구하기 - GetManyPath(width, height)

        /// <summary>
        /// Many 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>Many 패스</returns>
        private static GraphicsPath GetManyPath(float width, float height)
        {
            GraphicsPath graphicsPath = new GraphicsPath();

            graphicsPath.AddLine(new PointF(0, height / 2), new PointF(width, 0     ));
            graphicsPath.AddLine(new PointF(0, height / 2), new PointF(width, height));

            return graphicsPath;
        }

        #endregion
        #region One Or Many 패스 구하기 - GetOneOrManyPath(width, height)

        /// <summary>
        /// One Or Many 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>One Or Many 패스</returns>
        private static GraphicsPath GetOneOrManyPath(float width, float height)
        {
            GraphicsPath graphicsPath = new GraphicsPath();

            graphicsPath.AddLine(new PointF(0, height / 2), new PointF(0    , 0     ));
            graphicsPath.AddLine(new PointF(0, height / 2), new PointF(0    , height));
            graphicsPath.AddLine(new PointF(0, height / 2), new PointF(width, 0     ));
            graphicsPath.AddLine(new PointF(0, height / 2), new PointF(width, height));

            return graphicsPath;
        }

        #endregion
        #region One 패스 구하기 - GetOnePath(width, height)

        /// <summary>
        /// One 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>One 패스</returns>
        private static GraphicsPath GetOnePath(float width, float height)
        {
            GraphicsPath graphicsPath = new GraphicsPath();

            graphicsPath.AddLine(new PointF(0, 0         ), new PointF(0    , height    ));
            graphicsPath.AddLine(new PointF(0, height / 2), new PointF(width, height / 2));

            return graphicsPath;
        }

        #endregion
        #region Many Optional 패스 구하기 - GetManyOptionalPath(width, height)

        /// <summary>
        /// Many Optional 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>Many Optional 패스</returns>
        private static GraphicsPath GetManyOptionalPath(float width, float height)
        {
            RectangleF   rectangle    = new RectangleF(0, 0, width / 2, height);
            GraphicsPath graphicsPath = new GraphicsPath();

            graphicsPath.AddArc(rectangle, 0  , 180);
            graphicsPath.AddArc(rectangle, 180, 180);

            graphicsPath.CloseFigure();

            graphicsPath.AddLine(new PointF(width / 2, height / 2), new PointF(width, 0     ));
            graphicsPath.AddLine(new PointF(width / 2, height / 2), new PointF(width, height));

            return graphicsPath;
        }

        #endregion
        #region One Optional 패스 구하기 - GetOneOptionalPath(width, height)

        /// <summary>
        /// One Optional 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>One Optional 패스</returns>
        private static GraphicsPath GetOneOptionalPath(float width, float height)
        {
            RectangleF   rectangle    = new RectangleF(0, 0, width / 2, height);
            GraphicsPath graphicsPath = new GraphicsPath();

            graphicsPath.AddArc(rectangle, 0  , 180);
            graphicsPath.AddArc(rectangle, 180, 180);

            graphicsPath.CloseFigure();

            graphicsPath.AddLine(new PointF(3 * width / 4, height / 2), new PointF(3 * width / 4, 0     ));
            graphicsPath.AddLine(new PointF(3 * width / 4, height / 2), new PointF(3 * width / 4, height));

            Matrix matrix = new Matrix();

            matrix.Translate(-width / 4, 0);

            graphicsPath.Transform(matrix);

            return graphicsPath;
        }

        #endregion
        #region 반쪽 화살표 패스 구하기 - GetHalfArrowPath(width, height)

        /// <summary>
        /// 반쪽 화살표 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>반쪽 화살표 패스</returns>
        private static GraphicsPath GetHalfArrowPath(float width, float height)
        {
            GraphicsPath graphicsPath = new GraphicsPath();
            PointF[]     pointArray   = { new PointF(0, 0), new PointF(width, height / 2), new PointF(0, height / 2) };

            graphicsPath.AddPolygon(pointArray);

            Matrix matrix = new Matrix();

            matrix.Translate(0, -height / 4);

            graphicsPath.Transform(matrix);

            return graphicsPath;
        }

        #endregion
        #region 개방형 반쪽 화살표 패스 구하기 - GetOpenedHalfArrowPath(width, height)

        /// <summary>
        /// 개방형 반쪽 화살표 패스 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>개방형 반쪽 화살표 패스</returns>
        private static GraphicsPath GetOpenedHalfArrowPath(float width, float height)
        {
            GraphicsPath graphicsPath = new GraphicsPath();

            graphicsPath.AddLine(new PointF(width, height/2), new PointF(0, 0));

            Matrix matrix = new Matrix();

            matrix.Translate(0, -height / 4);

            graphicsPath.Transform(matrix);

            return graphicsPath;
        } 

        #endregion

        #region 중심 포인트 구하기 - GetCenterPoint(rectangle)

        /// <summary>
        /// 중심 포인트 구하기
        /// </summary>
        /// <param name="rectangle">사각형</param>
        /// <returns>중심 포인트</returns>
        private static PointF GetCenterPoint(RectangleF rectangle)
        {
            return new PointF(rectangle.Left + rectangle.Width / 2, rectangle.Top + rectangle.Height / 2);
        }

        #endregion
        #region 중간 포인트 구하기 - GetMiddlePoint(point1, point2)

        /// <summary>
        /// 중간 포인트 구하기
        /// </summary>
        /// <param name="point1">포인트 1</param>
        /// <param name="point2">포인트 2</param>
        /// <returns>중간 포인트</returns>
        private static PointF GetMiddlePoint(PointF point1, PointF point2)
        {
            return new PointF((point1.X + point2.X) / 2, (point1.Y + point2.Y) / 2);
        }

        #endregion
    }
}